@@ -1,4 +1,4 @@
-# Copyright 2016-2020 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
+# Copyright 2016-2021 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
 # ReFrame Project Developers. See the top-level LICENSE file for details.
 #
 # SPDX-License-Identifier: BSD-3-Clause
@@ -27,13 +27,20 @@ from reframe.utility.versioning import VersionValidator
 
 
 def _register_test(cls, args=None):
+    '''Register the test.
+
+    Register the test with _rfm_use_params=True. This additional argument flags
+    this case to consume the parameter space. Otherwise, the regression test
+    parameters would simply be initialized to None.
+    '''
     def _instantiate(cls, args):
         if isinstance(args, collections.abc.Sequence):
-            return cls(*args)
+            return cls(*args, _rfm_use_params=True)
         elif isinstance(args, collections.abc.Mapping):
+            args['_rfm_use_params'] = True
             return cls(**args)
         elif args is None:
-            return cls()
+            return cls(_rfm_use_params=True)
 
     def _instantiate_all():
         ret = []
@@ -72,9 +79,13 @@ def _validate_test(cls):
         raise ReframeSyntaxError('the decorated class must be a '
                                  'subclass of RegressionTest')
 
+    if (cls.is_abstract()):
+        raise ValueError(f'decorated test ({cls.__qualname__!r}) is an'
+                         f' abstract test')
+
 
 def simple_test(cls):
-    '''Class decorator for registering parameterless tests with ReFrame.
+    '''Class decorator for registering tests with ReFrame.
 
     The decorated class must derive from
     :class:`reframe.core.pipeline.RegressionTest`.  This decorator is also
@@ -82,9 +93,11 @@ def simple_test(cls):
 
     .. versionadded:: 2.13
     '''
-
     _validate_test(cls)
-    _register_test(cls)
+
+    for _ in cls.param_space:
+        _register_test(cls)
+
     return cls
 
 
@@ -106,6 +119,11 @@ def parameterized_test(*inst):
     '''
     def _do_register(cls):
         _validate_test(cls)
+        if not cls.param_space.is_empty():
+            raise ValueError(
+                f'{cls.__qualname__!r} is already a parameterized test'
+            )
+
         for args in inst:
             _register_test(cls, args)
 
