@@ -1,18 +1,19 @@
-# Copyright 2016-2020 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
+# Copyright 2016-2021 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
 # ReFrame Project Developers. See the top-level LICENSE file for details.
 #
 # SPDX-License-Identifier: BSD-3-Clause
 
 import json
-import re
 
 import reframe.utility as utility
+import reframe.utility.jsonext as jsonext
 from reframe.core.backends import (getlauncher, getscheduler)
+from reframe.core.logging import getlogger
 from reframe.core.modules import ModulesSystem
 from reframe.core.environments import (Environment, ProgEnvironment)
 
 
-class SystemPartition:
+class SystemPartition(jsonext.JSONSerializable):
     '''A representation of a system partition inside ReFrame.
 
     .. warning::
@@ -22,6 +23,7 @@ class SystemPartition:
     def __init__(self, parent, name, sched_type, launcher_type,
                  descr, access, container_environs, resources,
                  local_env, environs, max_jobs):
+        getlogger().debug(f'Initializing system partition {name!r}')
         self._parent_system = parent
         self._name = name
         self._sched_type = sched_type
@@ -155,7 +157,7 @@ class SystemPartition:
            Please use :attr:`launcher_type` instead.
         '''
 
-        from reframe.core.exceptions import user_deprecation_warning
+        from reframe.core.warnings import user_deprecation_warning
 
         user_deprecation_warning("the 'launcher' attribute is deprecated; "
                                  "please use 'launcher_type' instead")
@@ -235,7 +237,7 @@ class SystemPartition:
         return json.dumps(self.json(), indent=2)
 
 
-class System:
+class System(jsonext.JSONSerializable):
     '''A representation of a system inside ReFrame.
 
     .. warning::
@@ -245,6 +247,7 @@ class System:
     def __init__(self, name, descr, hostnames, modules_system,
                  preload_env, prefix, outputdir,
                  resourcesdir, stagedir, partitions):
+        getlogger().debug(f'Initializing system {name!r}')
         self._name = name
         self._descr = descr
         self._hostnames = hostnames
