@@ -1,4 +1,4 @@
-# Copyright 2016-2020 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
+# Copyright 2016-2021 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
 # ReFrame Project Developers. See the top-level LICENSE file for details.
 #
 # SPDX-License-Identifier: BSD-3-Clause
@@ -16,7 +16,6 @@ import traceback
 
 import reframe
 import reframe.core.config as config
-import reframe.core.environments as env
 import reframe.core.exceptions as errors
 import reframe.core.logging as logging
 import reframe.core.runtime as runtime
@@ -24,13 +23,16 @@ import reframe.core.warnings as warnings
 import reframe.frontend.argparse as argparse
 import reframe.frontend.dependencies as dependencies
 import reframe.frontend.filters as filters
+import reframe.frontend.runreport as runreport
 import reframe.utility.jsonext as jsonext
 import reframe.utility.osext as osext
-from reframe.frontend.executors import Runner, generate_testcases
+
+
+from reframe.frontend.printer import PrettyPrinter
+from reframe.frontend.loader import RegressionCheckLoader
 from reframe.frontend.executors.policies import (SerialExecutionPolicy,
                                                  AsynchronousExecutionPolicy)
-from reframe.frontend.loader import RegressionCheckLoader
-from reframe.frontend.printer import PrettyPrinter
+from reframe.frontend.executors import Runner, generate_testcases
 
 
 def format_check(check, check_deps, detailed=False):
@@ -57,7 +59,8 @@ def format_check(check, check_deps, detailed=False):
         return f'- {check.name} (found in {location!r})'
 
     if check.num_tasks > 0:
-        node_alloc_scheme = f'standard ({check.num_tasks} task(s))'
+        node_alloc_scheme = (f'standard ({check.num_tasks} task(s) -- '
+                             f'may be set differently in hooks)')
     elif check.num_tasks == 0:
         node_alloc_scheme = 'flexible'
     else:
@@ -119,23 +122,6 @@ def list_checks(testcases, printer, detailed=False):
     printer.info(f'Found {len(checks)} check(s)')
 
 
-def generate_report_filename(filepatt):
-    if '{sessionid}' not in filepatt:
-        return filepatt
-
-    search_patt = os.path.basename(filepatt).replace('{sessionid}', r'(\d+)')
-    new_id = -1
-    basedir = os.path.dirname(filepatt) or '.'
-    for filename in os.listdir(basedir):
-        match = re.match(search_patt, filename)
-        if match:
-            found_id = int(match.group(1))
-            new_id = max(found_id, new_id)
-
-    new_id += 1
-    return filepatt.format(sessionid=new_id)
-
-
 def logfiles_message():
     log_files = logging.log_files()
     msg = 'Log file(s) saved in: '
@@ -260,6 +246,10 @@ def main():
         help=('Select checks with at least one '
               'programming environment matching PATTERN')
     )
+    select_options.add_argument(
+        '--failed', action='store_true',
+        help="Select failed test cases (only when '--restore-session' is used)"
+    )
     select_options.add_argument(
         '--gpu-only', action='store_true',
         help='Select only GPU checks'
@@ -326,6 +316,11 @@ def main():
         help='Set the maximum number of times a failed regression test '
              'may be retried (default: 0)'
     )
+    run_options.add_argument(
+        '--restore-session', action='store', nargs='?', const='',
+        metavar='REPORT',
+        help='Restore a testing session from REPORT file'
+    )
     run_options.add_argument(
         '--flex-alloc-nodes', action='store',
         dest='flex_alloc_nodes', metavar='{all|STATE|NUM}', default=None,
@@ -586,10 +581,53 @@ def main():
     printer.debug(format_env(options.env_vars))
 
     # Setup the check loader
+    if options.restore_session is not None:
+        # We need to load the failed checks only from a report
+        if options.restore_session:
+            filename = options.restore_session
+        else:
+            filename = runreport.next_report_filename(
+                osext.expandvars(site_config.get('general/0/report_file')),
+                new=False
+            )
+
+        report = runreport.load_report(filename)
+        check_search_path = list(report.slice('filename', unique=True))
+        check_search_recursive = False
+
+        # If `-c` or `-R` are passed explicitly outside the configuration
+        # file, override the values set from the report file
+        if site_config.is_sticky_option('general/check_search_path'):
+            printer.warning(
+                'Ignoring check search path set in the report file: '
+                'search path set explicitly in the command-line or '
+                'the environment'
+            )
+            check_search_path = site_config.get(
+                'general/0/check_search_path'
+            )
+
+        if site_config.is_sticky_option('general/check_search_recursive'):
+            printer.warning(
+                'Ignoring check search recursive option from the report file: '
+                'option set explicitly in the command-line or the environment'
+            )
+            check_search_recursive = site_config.get(
+                'general/0/check_search_recursive'
+            )
+
+    else:
+        check_search_recursive = site_config.get(
+            'general/0/check_search_recursive'
+        )
+        check_search_path = site_config.get('general/0/check_search_path')
+
     loader = RegressionCheckLoader(
-        load_path=site_config.get('general/0/check_search_path'),
-        recurse=site_config.get('general/0/check_search_recursive'),
-        ignore_conflicts=site_config.get('general/0/ignore_check_conflicts')
+        load_path=check_search_path,
+        recurse=check_search_recursive,
+        ignore_conflicts=site_config.get(
+            'general/0/ignore_check_conflicts'
+        )
     )
 
     def print_infoline(param, value):
@@ -599,7 +637,7 @@ def main():
     session_info = {
         'cmdline': ' '.join(sys.argv),
         'config_file': rt.site_config.filename,
-        'data_version': '1.1',
+        'data_version': runreport.DATA_VERSION,
         'hostname': socket.gethostname(),
         'prefix_output': rt.output_prefix,
         'prefix_stage': rt.stage_prefix,
@@ -683,6 +721,34 @@ def main():
         elif options.cpu_only:
             testcases = filter(filters.have_cpu_only(), testcases)
 
+        testcases = list(testcases)
+        printer.verbose(
+            f'Filtering test cases(s) by other attributes: '
+            f'{len(testcases)} remaining'
+        )
+
+        # Filter in failed cases
+        if options.failed:
+            if options.restore_session is None:
+                printer.error(
+                    "the option '--failed' can only be used "
+                    "in combination with the '--restore-session' option"
+                )
+                sys.exit(1)
+
+            def _case_failed(t):
+                rec = report.case(*t)
+                if rec and rec['result'] == 'failure':
+                    return True
+                else:
+                    return False
+
+            testcases = list(filter(_case_failed, testcases))
+            printer.verbose(
+                f'Filtering successful test case(s): '
+                f'{len(testcases)} remaining'
+            )
+
         # Prepare for running
         printer.debug('Building and validating the full test DAG')
         testgraph, skipped_cases = dependencies.build_deps(testcases_all)
@@ -697,12 +763,22 @@ def main():
         dependencies.validate_deps(testgraph)
         printer.debug('Full test DAG:')
         printer.debug(dependencies.format_deps(testgraph))
+
+        restored_cases = []
         if len(testcases) != len(testcases_all):
-            testgraph = dependencies.prune_deps(testgraph, testcases)
+            testgraph = dependencies.prune_deps(
+                testgraph, testcases,
+                max_depth=1 if options.restore_session is not None else None
+            )
             printer.debug('Pruned test DAG')
             printer.debug(dependencies.format_deps(testgraph))
+            if options.restore_session is not None:
+                testgraph, restored_cases = report.restore_dangling(testgraph)
 
-        testcases = dependencies.toposort(testgraph)
+        testcases = dependencies.toposort(
+            testgraph,
+            is_subgraph=options.restore_session is not None
+        )
         printer.verbose(f'Final number of test cases: {len(testcases)}')
 
         # Disable hooks
@@ -738,34 +814,47 @@ def main():
             printer.debug(str(e))
             raise
 
+        def module_use(*paths):
+            try:
+                rt.modules_system.searchpath_add(*paths)
+            except errors.EnvironError as e:
+                printer.warning(f'could not add module paths correctly')
+                printer.debug(str(e))
+
+        def module_unuse(*paths):
+            try:
+                rt.modules_system.searchpath_remove(*paths)
+            except errors.EnvironError as e:
+                printer.warning(f'could not remove module paths correctly')
+                printer.debug(str(e))
+
         printer.debug('(Un)using module paths from command line')
+        module_paths = {}
         for d in options.module_paths:
             if d.startswith('-'):
-                try:
-                    rt.modules_system.searchpath_remove(d[1:])
-                except errors.EnvironError as e:
-                    printer.warning(
-                        f'could not remove module path {d} correctly; '
-                        f'skipping...'
-                    )
-                    printer.verbose(str(e))
+                module_paths.setdefault('-', [])
+                module_paths['-'].append(d[1:])
             elif d.startswith('+'):
-                try:
-                    rt.modules_system.searchpath_add(d[1:])
-                except errors.EnvironError as e:
-                    printer.warning(
-                        f'could not add module path {d} correctly; '
-                        f'skipping...'
-                    )
-                    printer.verbose(str(e))
+                module_paths.setdefault('+', [])
+                module_paths['+'].append(d[1:])
             else:
-                # Here we make sure that we don't try to remove an empty path
-                # from the searchpath
+                module_paths.setdefault('x', [])
+                module_paths['x'].append(d)
+
+        for op, paths in module_paths.items():
+            if op == '+':
+                module_use(*paths)
+            elif op == '-':
+                module_unuse(*paths)
+            else:
+                # First empty the current module path in a portable way
                 searchpath = [p for p in rt.modules_system.searchpath if p]
                 if searchpath:
                     rt.modules_system.searchpath_remove(*searchpath)
 
-                rt.modules_system.searchpath_add(d)
+                # Treat `A:B` syntax as well in this case
+                paths = itertools.chain(*(p.split(':') for p in paths))
+                module_use(*paths)
 
         printer.debug('Loading user modules from command line')
         for m in site_config.get('general/0/user_modules'):
@@ -835,7 +924,7 @@ def main():
             session_info['time_start'] = time.strftime(
                 '%FT%T%z', time.localtime(time_start),
             )
-            runner.runall(testcases)
+            runner.runall(testcases, restored_cases)
         finally:
             time_end = time.time()
             session_info['time_end'] = time.strftime(
@@ -874,9 +963,14 @@ def main():
             })
             json_report = {
                 'session_info': session_info,
-                'runs': run_stats
+                'runs': run_stats,
+                'restored_cases': []
             }
-            report_file = generate_report_filename(report_file)
+            if options.restore_session is not None:
+                for c in restored_cases:
+                    json_report['restored_cases'].append(report.case(*c))
+
+            report_file = runreport.next_report_filename(report_file)
             try:
                 with open(report_file, 'w') as fp:
                     jsonext.dump(json_report, fp, indent=2)
