@@ -1,25 +1,27 @@
-# Copyright 2016-2020 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
+# Copyright 2016-2021 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
 # ReFrame Project Developers. See the top-level LICENSE file for details.
 #
 # SPDX-License-Identifier: BSD-3-Clause
 
 import abc
 import copy
+import os
 import signal
 import sys
 import time
 import weakref
 
-import reframe.core.environments as env
 import reframe.core.logging as logging
 import reframe.core.runtime as runtime
 import reframe.frontend.dependencies as dependencies
+import reframe.utility.jsonext as jsonext
 from reframe.core.exceptions import (AbortTaskError, JobNotStartedError,
                                      ReframeForceExitError, TaskExit)
 from reframe.core.schedulers.local import LocalJobScheduler
 from reframe.frontend.printer import PrettyPrinter
 from reframe.frontend.statistics import TestStats
 
+
 ABORT_REASONS = (KeyboardInterrupt, ReframeForceExitError, AssertionError)
 
 
@@ -29,12 +31,12 @@ class TestCase:
     '''
 
     def __init__(self, check, partition, environ):
-        self.__check_orig = check
-        self.__check = copy.deepcopy(check)
-        self.__partition = copy.deepcopy(partition)
-        self.__environ = copy.deepcopy(environ)
-        self.__check._case = weakref.ref(self)
-        self.__deps = []
+        self._check_orig = check
+        self._check = copy.deepcopy(check)
+        self._partition = copy.deepcopy(partition)
+        self._environ = copy.deepcopy(environ)
+        self._check._case = weakref.ref(self)
+        self._deps = []
 
         # Incoming dependencies
         self.in_degree = 0
@@ -42,7 +44,7 @@ class TestCase:
     def __iter__(self):
         # Allow unpacking a test case with a single liner:
         #       c, p, e = case
-        return iter([self.__check, self.__partition, self.__environ])
+        return iter([self._check, self._partition, self._environ])
 
     def __hash__(self):
         return (hash(self.check.name) ^
@@ -63,19 +65,19 @@ class TestCase:
 
     @property
     def check(self):
-        return self.__check
+        return self._check
 
     @property
     def partition(self):
-        return self.__partition
+        return self._partition
 
     @property
     def environ(self):
-        return self.__environ
+        return self._environ
 
     @property
     def deps(self):
-        return self.__deps
+        return self._deps
 
     @property
     def num_dependents(self):
@@ -83,7 +85,7 @@ class TestCase:
 
     def clone(self):
         # Return a fresh clone, i.e., one based on the original check
-        return TestCase(self.__check_orig, self.__partition, self.__environ)
+        return TestCase(self._check_orig, self._partition, self._environ)
 
 
 def generate_testcases(checks,
@@ -287,6 +289,13 @@ class RegressionTask:
         self._safe_call(self.check.performance)
 
     def finalize(self):
+        try:
+            jsonfile = os.path.join(self.check.stagedir, '.rfm_testcase.json')
+            with open(jsonfile, 'w') as fp:
+                jsonext.dump(self.check, fp, indent=2)
+        except OSError as e:
+            self._printer.warning(f'could not dump test case {self.case}: {e}')
+
         self._current_stage = 'finalize'
         self._notify_listeners('on_task_success')
 
@@ -366,7 +375,7 @@ class Runner:
     def stats(self):
         return self._stats
 
-    def runall(self, testcases):
+    def runall(self, testcases, restored_cases=None):
         num_checks = len({tc.check.name for tc in testcases})
         self._printer.separator('short double line',
                                 'Running %d check(s)' % num_checks)
@@ -375,7 +384,8 @@ class Runner:
         try:
             self._runall(testcases)
             if self._max_retries:
-                self._retry_failed(testcases)
+                restored_cases = restored_cases or []
+                self._retry_failed(testcases + restored_cases)
 
         finally:
             # Print the summary line
