@@ -1,22 +1,24 @@
-# Copyright 2016-2020 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
+# Copyright 2016-2021 Swiss National Supercomputing Centre (CSCS/ETH Zurich)
 # ReFrame Project Developers. See the top-level LICENSE file for details.
 #
 # SPDX-License-Identifier: BSD-3-Clause
 
 import contextlib
+import functools
 import itertools
-import math
 import sys
 import time
 
-from datetime import datetime
-
 from reframe.core.exceptions import (TaskDependencyError, TaskExit)
 from reframe.core.logging import getlogger
 from reframe.frontend.executors import (ExecutionPolicy, RegressionTask,
                                         TaskEventListener, ABORT_REASONS)
 
 
+def countall(d):
+    return functools.reduce(lambda l, r: l + len(r), d.values(), 0)
+
+
 def _cleanup_all(tasks, *args, **kwargs):
     for task in tasks:
         if task.ref_count == 0:
@@ -27,10 +29,50 @@ def _cleanup_all(tasks, *args, **kwargs):
     tasks[:] = [t for t in tasks if t.ref_count]
 
 
+class _PollController:
+    SLEEP_MIN = 0.1
+    SLEEP_MAX = 10
+    SLEEP_INC_RATE = 1.1
+
+    def __init__(self):
+        self._num_polls = 0
+        self._num_tasks = 0
+        self._sleep_duration = None
+        self._t_init = None
+
+    def running_tasks(self, num_tasks):
+        if self._sleep_duration is None:
+            self._sleep_duration = self.SLEEP_MIN
+
+        if self._num_polls == 0:
+            self._t_init = time.time()
+        else:
+            if self._num_tasks != num_tasks:
+                self._sleep_duration = self.SLEEP_MIN
+            else:
+                self._sleep_duration = min(
+                    self._sleep_duration*self.SLEEP_INC_RATE, self.SLEEP_MAX
+                )
+
+        self._num_tasks = num_tasks
+        return self
+
+    def snooze(self):
+        t_elapsed = time.time() - self._t_init
+        self._num_polls += 1
+        getlogger().debug2(
+            f'Poll rate control: sleeping for {self._sleep_duration}s '
+            f'(current poll rate: {self._num_polls/t_elapsed} polls/s)'
+        )
+        time.sleep(self._sleep_duration)
+
+
 class SerialExecutionPolicy(ExecutionPolicy, TaskEventListener):
     def __init__(self):
         super().__init__()
 
+        self._pollctl = _PollController()
+
         # Index tasks by test cases
         self._task_index = {}
 
@@ -51,22 +93,35 @@ class SerialExecutionPolicy(ExecutionPolicy, TaskEventListener):
         self.stats.add_task(task)
         try:
             # Do not run test if any of its dependencies has failed
-            if any(self._task_index[c].failed for c in case.deps):
+            # NOTE: Restored dependencies are not in the task_index
+            if any(self._task_index[c].failed
+                   for c in case.deps if c in self._task_index):
                 raise TaskDependencyError('dependencies failed')
 
-            task.setup(partition, environ,
+            partname = task.testcase.partition.fullname
+            task.setup(task.testcase.partition,
+                       task.testcase.environ,
                        sched_flex_alloc_nodes=self.sched_flex_alloc_nodes,
-                       sched_account=self.sched_account,
-                       sched_partition=self.sched_partition,
-                       sched_reservation=self.sched_reservation,
-                       sched_nodelist=self.sched_nodelist,
-                       sched_exclude_nodelist=self.sched_exclude_nodelist,
                        sched_options=self.sched_options)
 
             task.compile()
             task.compile_wait()
             task.run()
-            task.wait()
+
+            # Pick the right scheduler
+            if task.check.local:
+                sched = self.local_scheduler
+            else:
+                sched = partition.scheduler
+
+            while True:
+                sched.poll(task.check.job)
+                if task.run_complete():
+                    break
+
+                self._pollctl.running_tasks(1).snooze()
+
+            task.run_wait()
             if not self.skip_sanity_check:
                 task.sanity()
 
@@ -109,7 +164,9 @@ class SerialExecutionPolicy(ExecutionPolicy, TaskEventListener):
                                          'sanity',
                                          'performance',
                                          'total'])
-        getlogger().verbose(f"==> {timings}")
+        getlogger().info(f'==> test failed during {task.failed_stage!r}: '
+                         f'test staged in {task.check.stagedir!r}')
+        getlogger().verbose(f'==> {timings}')
 
     def on_task_success(self, task):
         timings = task.pipeline_timings(['compile_complete',
@@ -123,10 +180,13 @@ class SerialExecutionPolicy(ExecutionPolicy, TaskEventListener):
                                          'sanity',
                                          'performance',
                                          'total'])
-        getlogger().verbose(f"==> {timings}")
-        # update reference count of dependencies
+        getlogger().verbose(f'==> {timings}')
+
+        # Update reference count of dependencies
         for c in task.testcase.deps:
-            self._task_index[c].ref_count -= 1
+            # NOTE: Restored dependencies are not in the task_index
+            if c in self._task_index:
+                self._task_index[c].ref_count -= 1
 
         _cleanup_all(self._retired_tasks, not self.keep_stage_files)
 
@@ -135,48 +195,18 @@ class SerialExecutionPolicy(ExecutionPolicy, TaskEventListener):
         _cleanup_all(self._retired_tasks, not self.keep_stage_files)
 
 
-class PollRateFunction:
-    def __init__(self, min_rate, decay_time):
-        self._min_rate = min_rate
-        self._decay = decay_time
-        self._thres = 0.05
-
-        # decay function parameters
-        self._a = None
-        self._b = None
-        self._c = None
-
-    def _init_poll_fn(self, init_rate):
-        self._init_rate = init_rate
-        self._b = self._min_rate
-        log_arg = (init_rate - self._b) / (self._thres*self._b)
-        if log_arg < sys.float_info.min:
-            self._a = 0.0
-            self._c = 0.0
-        else:
-            self._a = init_rate - self._b
-            self._c = math.log(self._a / (self._thres*self._b)) / self._decay
-
-        getlogger().debug('rate equation: %.3f*exp(-%.3f*x)+%.3f' %
-                          (self._a, self._c, self._b))
-
-    def __call__(self, x, init_rate):
-        if self._a is None:
-            self._init_poll_fn(init_rate)
-
-        return self._a*math.exp(-self._c*x) + self._b
-
-
 class AsynchronousExecutionPolicy(ExecutionPolicy, TaskEventListener):
     def __init__(self):
 
         super().__init__()
 
+        self._pollctl = _PollController()
+
         # Index tasks by test cases
         self._task_index = {}
 
-        # All currently running tasks
-        self._running_tasks = []
+        # All currently running tasks per partition
+        self._running_tasks = {}
 
         # Tasks that need to be finalized
         self._completed_tasks = []
@@ -184,9 +214,6 @@ class AsynchronousExecutionPolicy(ExecutionPolicy, TaskEventListener):
         # Retired tasks that need to be cleaned up
         self._retired_tasks = []
 
-        # Counts of running tasks per partition
-        self._running_tasks_counts = {}
-
         # Ready tasks to be executed per partition
         self._ready_tasks = {}
 
@@ -196,26 +223,31 @@ class AsynchronousExecutionPolicy(ExecutionPolicy, TaskEventListener):
         # Job limit per partition
         self._max_jobs = {}
 
+        # Keep a reference to all the partitions
+        self._partitions = set()
+
         self.task_listeners.append(self)
 
     def _remove_from_running(self, task):
-        getlogger().debug(
-            'removing task from running list: %s' % task.check.info()
+        getlogger().debug2(
+            f'Removing task from the running list: {task.testcase}'
         )
         try:
-            self._running_tasks.remove(task)
-        except ValueError:
-            getlogger().debug('not in running tasks')
-            pass
-        else:
             partname = task.check.current_partition.fullname
-            self._running_tasks_counts[partname] -= 1
+            self._running_tasks[partname].remove(task)
+        except (ValueError, AttributeError, KeyError):
+            getlogger().debug2('Task was not running')
+            pass
 
     def deps_failed(self, task):
-        return any(self._task_index[c].failed for c in task.testcase.deps)
+        # NOTE: Restored dependencies are not in the task_index
+        return any(self._task_index[c].failed
+                   for c in task.testcase.deps if c in self._task_index)
 
     def deps_succeeded(self, task):
-        return all(self._task_index[c].succeeded for c in task.testcase.deps)
+        # NOTE: Restored dependencies are not in the task_index
+        return all(self._task_index[c].succeeded
+                   for c in task.testcase.deps if c in self._task_index)
 
     def on_task_setup(self, task):
         partname = task.check.current_partition.fullname
@@ -223,8 +255,7 @@ class AsynchronousExecutionPolicy(ExecutionPolicy, TaskEventListener):
 
     def on_task_run(self, task):
         partname = task.check.current_partition.fullname
-        self._running_tasks_counts[partname] += 1
-        self._running_tasks.append(task)
+        self._running_tasks[partname].append(task)
 
     def on_task_failure(self, task):
         msg = f'{task.check.info()} [{task.pipeline_timings_basic()}]'
@@ -234,20 +265,29 @@ class AsynchronousExecutionPolicy(ExecutionPolicy, TaskEventListener):
             self._remove_from_running(task)
             self.printer.status('FAIL', msg, just='right')
 
-        getlogger().verbose(f"==> {task.pipeline_timings_all()}")
+        stagedir = task.check.stagedir
+        if not stagedir:
+            stagedir = '<not available>'
+
+        getlogger().info(f'==> test failed during {task.failed_stage!r}: '
+                         f'test staged in {stagedir!r}')
+        getlogger().verbose(f'==> timings: {task.pipeline_timings_all()}')
 
     def on_task_success(self, task):
         msg = f'{task.check.info()} [{task.pipeline_timings_basic()}]'
         self.printer.status('OK', msg, just='right')
-        getlogger().verbose(f"==> {task.pipeline_timings_all()}")
-        # update reference count of dependencies
+        getlogger().verbose(f'==> timings: {task.pipeline_timings_all()}')
+
+        # Update reference count of dependencies
         for c in task.testcase.deps:
-            self._task_index[c].ref_count -= 1
+            # NOTE: Restored dependencies are not in the task_index
+            if c in self._task_index:
+                self._task_index[c].ref_count -= 1
 
         self._retired_tasks.append(task)
 
     def on_task_exit(self, task):
-        task.wait()
+        task.run_wait()
         self._remove_from_running(task)
         self._completed_tasks.append(task)
 
@@ -257,11 +297,6 @@ class AsynchronousExecutionPolicy(ExecutionPolicy, TaskEventListener):
                 task.setup(task.testcase.partition,
                            task.testcase.environ,
                            sched_flex_alloc_nodes=self.sched_flex_alloc_nodes,
-                           sched_account=self.sched_account,
-                           sched_partition=self.sched_partition,
-                           sched_reservation=self.sched_reservation,
-                           sched_nodelist=self.sched_nodelist,
-                           sched_exclude_nodelist=self.sched_exclude_nodelist,
                            sched_options=self.sched_options)
             except TaskExit:
                 return False
@@ -278,9 +313,10 @@ class AsynchronousExecutionPolicy(ExecutionPolicy, TaskEventListener):
     def runcase(self, case):
         super().runcase(case)
         check, partition, environ = case
+        self._partitions.add(partition)
 
         # Set partition-based counters, if not set already
-        self._running_tasks_counts.setdefault(partition.fullname, 0)
+        self._running_tasks.setdefault(partition.fullname, [])
         self._ready_tasks.setdefault(partition.fullname, [])
         self._max_jobs.setdefault(partition.fullname, partition.max_jobs)
 
@@ -304,13 +340,15 @@ class AsynchronousExecutionPolicy(ExecutionPolicy, TaskEventListener):
 
                 return
 
-            if self._running_tasks_counts[partname] >= partition.max_jobs:
+            if len(self._running_tasks[partname]) >= partition.max_jobs:
                 # Make sure that we still exceeded the job limit
-                getlogger().debug('reached job limit (%s) for partition %s' %
-                                  (partition.max_jobs, partname))
+                getlogger().debug2(
+                    f'Reached concurrency limit for partition {partname!r}: '
+                    f'{partition.max_jobs} job(s)'
+                )
                 self._poll_tasks()
 
-            if self._running_tasks_counts[partname] < partition.max_jobs:
+            if len(self._running_tasks[partname]) < partition.max_jobs:
                 # Task was put in _ready_tasks during setup
                 self._ready_tasks[partname].pop()
                 self._reschedule(task)
@@ -333,10 +371,32 @@ class AsynchronousExecutionPolicy(ExecutionPolicy, TaskEventListener):
 
     def _poll_tasks(self):
         '''Update the counts of running checks per partition.'''
-        getlogger().debug('updating counts for running test cases')
-        getlogger().debug('polling %s task(s)' % len(self._running_tasks))
-        for t in self._running_tasks:
-            t.poll()
+
+        def split_jobs(tasks):
+            '''Split jobs into forced local and normal ones.'''
+            forced_local = []
+            normal = []
+            for t in tasks:
+                if t.check.local:
+                    forced_local.append(t.check.job)
+                else:
+                    normal.append(t.check.job)
+
+            return forced_local, normal
+
+        for part in self._partitions:
+            partname = part.fullname
+            num_tasks = len(self._running_tasks[partname])
+            getlogger().debug2(f'Polling {num_tasks} task(s) in {partname!r}')
+            forced_local_jobs, part_jobs = split_jobs(
+                self._running_tasks[partname]
+            )
+            part.scheduler.poll(*part_jobs)
+            self.local_scheduler.poll(*forced_local_jobs)
+
+            # Trigger notifications for finished jobs
+            for t in self._running_tasks[partname]:
+                t.run_complete()
 
     def _setup_all(self):
         still_waiting = []
@@ -347,18 +407,19 @@ class AsynchronousExecutionPolicy(ExecutionPolicy, TaskEventListener):
         self._waiting_tasks[:] = still_waiting
 
     def _finalize_all(self):
-        getlogger().debug('finalizing tasks: %s', len(self._completed_tasks))
+        getlogger().debug2(f'Finalizing {len(self._completed_tasks)} task(s)')
         while True:
             try:
                 task = self._completed_tasks.pop()
             except IndexError:
                 break
 
-            getlogger().debug('finalizing task: %s' % task.check.info())
+            getlogger().debug2(f'Finalizing task {task.testcase}')
             with contextlib.suppress(TaskExit):
                 self._finalize_task(task)
 
     def _finalize_task(self, task):
+        getlogger().debug2(f'Finalizing task {task.testcase}')
         if not self.skip_sanity_check:
             task.sanity()
 
@@ -369,33 +430,29 @@ class AsynchronousExecutionPolicy(ExecutionPolicy, TaskEventListener):
 
     def _failall(self, cause):
         '''Mark all tests as failures'''
-        try:
-            while True:
-                self._running_tasks.pop().abort(cause)
-        except IndexError:
-            pass
+        getlogger().debug2(f'Aborting all tasks due to {type(cause).__name__}')
+        for task in list(itertools.chain(*self._running_tasks.values())):
+            task.abort(cause)
 
+        self._running_tasks = {}
         for ready_list in self._ready_tasks.values():
-            getlogger().debug('ready list size: %s' % len(ready_list))
             for task in ready_list:
                 task.abort(cause)
 
         for task in itertools.chain(self._waiting_tasks,
-                                    self._retired_tasks,
                                     self._completed_tasks):
             task.abort(cause)
 
     def _reschedule(self, task):
-        getlogger().debug('scheduling test case for running')
-
+        getlogger().debug2(f'Scheduling test case {task.testcase} for running')
         task.compile()
         task.compile_wait()
         task.run()
 
     def _reschedule_all(self):
-        for partname, num_jobs in self._running_tasks_counts.items():
-            assert(num_jobs >= 0)
-            num_empty_slots = self._max_jobs[partname] - num_jobs
+        for partname, tasks in self._running_tasks.items():
+            num_tasks = len(tasks)
+            num_empty_slots = self._max_jobs[partname] - num_tasks
             num_rescheduled = 0
             for _ in range(num_empty_slots):
                 try:
@@ -407,36 +464,32 @@ class AsynchronousExecutionPolicy(ExecutionPolicy, TaskEventListener):
                 num_rescheduled += 1
 
             if num_rescheduled:
-                getlogger().debug('rescheduled %s job(s) on %s' %
-                                  (num_rescheduled, partname))
+                getlogger().debug2(
+                    f'Rescheduled {num_rescheduled} job(s) on {partname!r}'
+                )
 
     def exit(self):
         self.printer.separator('short single line',
                                'waiting for spawned checks to finish')
-        pollrate = PollRateFunction(0.2, 60)
-        num_polls = 0
-        t_start = datetime.now()
-        while (self._running_tasks or self._waiting_tasks):
-            getlogger().debug('running tasks: %s' % len(self._running_tasks))
-            num_polls += len(self._running_tasks)
+        while (countall(self._running_tasks) or self._waiting_tasks or
+               self._completed_tasks or countall(self._ready_tasks)):
+            getlogger().debug2(f'Running tasks: '
+                               f'{countall(self._running_tasks)}')
             try:
                 self._poll_tasks()
+
+                # We count running tasks just after polling in order to check
+                # more reliably that the state has changed, so that we
+                # decrease the sleep time. Otherwise if the number of tasks
+                # rescheduled was the as the number of tasks retired, the
+                # sleep time would be increased.
+                num_running = countall(self._running_tasks)
                 self._finalize_all()
                 self._setup_all()
                 self._reschedule_all()
                 _cleanup_all(self._retired_tasks, not self.keep_stage_files)
-                t_elapsed = (datetime.now() - t_start).total_seconds()
-                real_rate = num_polls / t_elapsed
-                getlogger().debug(
-                    'polling rate (real): %.3f polls/sec' % real_rate)
-
-                if len(self._running_tasks):
-                    desired_rate = pollrate(t_elapsed, real_rate)
-                    getlogger().debug(
-                        'polling rate (desired): %.3f' % desired_rate)
-                    t = len(self._running_tasks) / desired_rate
-                    getlogger().debug('sleeping: %.3fs' % t)
-                    time.sleep(t)
+                if num_running:
+                    self._pollctl.running_tasks(num_running).snooze()
 
             except TaskExit:
                 with contextlib.suppress(TaskExit):
